#ifndef Q_CC_BOR
# if defined(UNDER_CE) && UNDER_CE >= 400
#  include <winbase.h>
# else
#  include <winver.h>
# endif
#endif

#include <config.h>

VS_VERSION_INFO VERSIONINFO
  FILEVERSION VERSION_COMMA
  PRODUCTVERSION VERSION_COMMA
  FILEFLAGSMASK VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS VS_FF_DEBUG
#else
  FILEFLAGS 0x0L
#endif
  FILEOS VOS_NT_WINDOWS32
  FILETYPE VFT_DLL
  FILESUBTYPE VFT_UNKNOWN
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904E4"
		BEGIN
			VALUE "CompanyName", "RIA"
			VALUE "FileDescription", "libdigidoc"
			VALUE "FileVersion", VERSION
			VALUE "InternalName", "libdigidoc"
			VALUE "LegalCopyright", "Copyright (C) 2009-2018 Estonian Information System Authority"
			VALUE "OriginalFilename", "libdigidoc.dll"
			VALUE "ProductName", "libdigidoc"
			VALUE "ProductVersion", VERSION
		END
	END
	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x425, 1200
	END
END
/* End of Version info */
