//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#define APSTUDIO_HIDDEN_SYMBOLS
#include "windows.h"
#undef APSTUDIO_HIDDEN_SYMBOLS
#include "dlgunit.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Unknown language: 0x25, 0x1 resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ETI)
#ifdef _WIN32
LANGUAGE 0x25, 0x1
#pragma code_page(1257)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_DLG_UNIT DIALOGEX 0, 0, 300, 155
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "DigiDoc"
FONT 8, "MS Sans Serif", 0, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "&OK",IDOK,240,136,50,14
    GROUPBOX        "Downloads:",IDC_GROUPBOX,2,2,296,126
    PUSHBUTTON      "&Cancel",IDCANCEL,180,136,50,14
    CONTROL         "List1",IDC_LISTVIEW,"SysListView32",LVS_REPORT | 
                    LVS_SINGLESEL | LVS_SHOWSELALWAYS | LVS_AUTOARRANGE | 
                    WS_BORDER | WS_TABSTOP,6,12,288,110,WS_EX_CLIENTEDGE
END


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#define APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "#include ""windows.h""\r\n"
    "#undef APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "#include ""dlgunit.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 2,7,1,56
 PRODUCTVERSION 2,7,1,56
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "042504b0"
        BEGIN
            VALUE "Comments", "\0"
            VALUE "CompanyName", "AS Sertifitseerimiskeskus\0"
            VALUE "FileDescription", "DigiDocLib\0"
            VALUE "FileVersion", "2.7.1.56\0"
            VALUE "InternalName", "DigiDocLib\0"
            VALUE "LegalCopyright", "Copyright © 2010\0"
            VALUE "LegalTrademarks", "\0"
            VALUE "OriginalFilename", "DigiDocLib.dll\0"
            VALUE "PrivateBuild", "\0"
            VALUE "ProductName", "AS Sertifitseerimiskeskus DigiDocLib\0"
            VALUE "ProductVersion", "2.7.1.56\0"
            VALUE "SpecialBuild", "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x425, 1200
    END
END

#endif    // !_MAC

#endif    // Unknown language: 0x25, 0x1 resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

